import abc
from typing import List

import httpx
from httpx._types import ProxiesTypes

from petfinder.models import AnimalQuery


class BaseClient(abc.ABC):
    _base_url: str
    _api_secret: str
    _api_key: str
    _authorization_headers: dict
    _proxies: ProxiesTypes

    def __init__(
        self,
        *,
        api_secret: str,
        api_key: str,
        proxies: ProxiesTypes = None,
        base_url: str = "https://api.petfinder.com/v2",
    ) -> None:
        self._api_secret = api_secret
        self._api_key = api_key
        self._proxies = proxies
        self._base_url = base_url
        self._refresh_access_token()

    def _refresh_access_token(self) -> None:
        """
        Retrieves a new access token and updates the auth headers.
        """
        response = httpx.post(
            f"{self._base_url}/oauth2/token",
            proxies=self._proxies,
            data={
                "grant_type": "client_credentials",
                "client_id": self._api_key,
                "client_secret": self._api_secret,
            },
        )
        response.raise_for_status()
        token = response.json()["access_token"]
        c = httpx.Client()
        c.close()
        self._authorization_headers = {"Authorization": f"Bearer {token}"}


class Client(BaseClient):
    def _client(self) -> httpx.Client:
        """
        Sets up the httpx client
        """
        return httpx.Client(
            base_url=self._base_url,
            headers=self._authorization_headers,
            proxies=self._proxies,
        )

    def _get(self, path: str, **params) -> dict:
        """
        Performs a GET request to the petfinder API.

        If the request fails due to an authentication error (401) we assume
        that the token has simply expired, so we refresh it and then call
        this function again.
        """
        try:
            with self._client() as client:
                response = client.get(path, params=params)
                response.raise_for_status()
                return response.json()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 401:
                self._refresh_access_token()
                return self._get(path, **params)
            else:
                raise e

    def get_animals(
        self, query: AnimalQuery = None, require_photos: bool = False
    ) -> List[dict]:
        """
        Retrieves a page of animals satisfying the given constraints
        """
        params = query.dict() if query else {}
        data = self._get(path="animals", **params)
        animals = data["animals"]
        return [a for a in animals if a.get("photos")] if require_photos else animals


class AsyncClient(BaseClient):
    def _client(self) -> httpx.AsyncClient:
        """
        Sets up the httpx async client
        """
        return httpx.AsyncClient(
            base_url=self._base_url,
            headers=self._authorization_headers,
            proxies=self._proxies,
        )

    async def _get(self, path: str, **params) -> dict:
        """
        Performs a GET request to the petfinder API.

        If the request fails due to an authentication error (401) we assume
        that the token has simply expired, so we refresh it and then call
        this function again.
        """
        try:
            async with self._client() as client:
                response = await client.get(path, params=params)
                response.raise_for_status()
                return response.json()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 401:
                self._refresh_access_token()
                return await self._get(path, **params)
            else:
                raise e

    async def get_animals(
        self, query: AnimalQuery = None, require_photos: bool = False
    ) -> List[dict]:
        """
        Retrieves a page of animals satisfying the given constraints
        """
        params = query.dict() if query else {}
        data = await self._get(path="animals", **params)
        animals = data["animals"]
        return [a for a in animals if a.get("photos")] if require_photos else animals



from enum import Enum
from typing import Optional

from pydantic import BaseModel
from pydantic.class_validators import root_validator
from pydantic.types import conint


class AnimalType(str, Enum):
    dog = "dog"
    cat = "cat"
    small_furry = "small-furry"
    bird = "bird"
    scales_fins_other = "scales-fins-other"
    barnyard = "barnyard"
    rabbit = "rabbit"
    horse = "horse"


class AnimalSize(str, Enum):
    small = "small"
    medium = "medium"
    large = "large"
    extra_large = "extra-large"


class AnimalGender(str, Enum):
    male = "male"
    female = "female"
    unknown = "unknown"


class AnimalCoat(str, Enum):
    short = "short"
    medium = "medium"
    long = "long"
    wire = "wire"
    hairless = "hairless"
    curly = "curly"


class AnimalStatus(str, Enum):
    adoptable = "adoptable"
    adopted = "adopted"
    found = "found"


class AnimalAge(str, Enum):
    baby = "baby"
    young = "young"
    adult = "adult"
    senior = "senior"


class SortType(str, Enum):
    recent = "recent"
    random = "random"


class AnimalQuery(BaseModel):
    type: Optional[AnimalType] = None
    age: Optional[AnimalAge] = None
    size: Optional[AnimalSize] = None
    gender: Optional[AnimalGender] = None
    breed: Optional[str] = None
    name: Optional[str] = None
    zip_code: Optional[str] = None
    distance: Optional[int] = None
    status: AnimalStatus = AnimalStatus.adoptable
    limit: conint(gt=0, le=100) = 100
    sort: SortType = SortType.random

    class Config(BaseModel.Config):
        use_enum_values = True

    def dict(self, *args, **kwargs):
        d = super().dict(exclude_none=True, by_alias=True)
        if "zip_code" in d:
            d["location"] = d.pop("zip_code")
        return d

    @root_validator
    def check_dependencies(cls, values: dict) -> dict:
        """
        Check through the values for field mismatches that result in a 400 (bad request) from the api.
        """
        if values.get("distance") and (values.get("zip_code") is None):
            raise Exception(
                "You cannot specify 'distance' without first providing a value for 'zip_code'"
            )
        return values



from setuptools import setup, find_packages


def readme():
    with open("README.md", "r") as infile:
        return infile.read()


classifiers = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
]

install_requires = [
    "httpx",
    "pydantic",
]

setup(
    name="petfinder-client",
    version="0.0.2",
    description="Petfinder API client",
    license="MIT",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords=["async"],
    author="Phillip Dupuis",
    author_email="phillip_dupuis@alumni.brown.edu",
    url="https://github.com/phillipdupuis/petfinder-client",
    install_requires=install_requires,
    packages=find_packages(),
    classifiers=classifiers,
)
